/**********************************************************************************
   Copyright(c) Catalyst Semiconductor, Inc., 2006. All rights reserved.

This code is provided as is without warranty or guarantee of any kind as its accuracy, 
completeness, operability, fitness for particular purpose, or any other warranty.

   File        :  IICslave.v
   Description :  Verilog model for CAT24XXXX I2C serial EEPROMs
		  (viewd as Verilog model for IIC protocol - slave side)

   Version     :  Rev1.01 
   Release     :  Sep, 2006
**********************************************************************************/

module IICslave( A0, A1, A2, SCL, SDA, WP, VCC, VSS);

input 		A2;	// Device Address Bit 2
input 		A1;	// Device Address Bit 1
input 		A0;	// Device Address Bit 0
inout 		SDA;	// Serial Data/Address
input		SCL;	// Serial Clock
input 		WP;	// Write Protect
input		VCC;
supply1		VCC;
input		VSS;
supply0		VSS;

`define B0	1'b0
`define B1	1'b1
`define N	1'b0
`define Y	1'b1

`define idle 			4'b0000
`define receivingAddrInfo 	4'b0001
`define receivingAddr 		4'b0010
`define receivingData 		4'b0011
`define sendingData		4'b0111
`define sendingNextData		4'b1000
`define waitingAck		4'b1001
`define waitingStop		4'b1010
`define internalWriting		4'b1011

`define nothingSpecial		2'b00
`define startReceived 		2'b01
`define ackReceived 		2'b10
`define stopReceived		2'b11

`define OUT			2'b00
`define HIGH			2'b01

`define write 			2'b00
`define read 			2'b01
`define noop			2'b11

`define immedAddrRead 		2'b00
`define selectiveRead 		2'b01
`define sequentialImmedRead 	2'b10
`define sequentialSelectRead 	2'b11

//-----------------------------------------------------------------------

parameter delta = 10;

//-----------------------------------------------------------------------

reg[7:0] 	addrInfoByteR;
reg[7:0] 	addrInfoByteW; 
reg[3:0] 	headerFirstHalf, reqHeaderFirstHalf;

reg [7:0] 	mem [0:32767];			// memory block (256K bits)
reg [7:0] 	data[0:100];			// buffer for data to write 

reg[8*10:1] 	chipName;				
reg [15:0] 	addrMax, addrBegProt, addrEndProt;
reg [15:0] 	Saddr;	
reg [2:0] 	addrDev,realAddrDev;
reg [2:0] 	extraAddrBits;
reg [1:0] 	nrExtraAddrBits;

reg [3:0] 	state,nextState,afterAckState,busState;	// State machine controller
reg [1:0]	condition;			// Special condition (START / STOP / ACK ) detection
reg [1:0]	operation, savOperation;	// Operation (Read/Write) 
reg [1:0] 	operationType;			// Operation type (certain kind of read or write)

reg [7:0]	tempReg;			// register to assemblate the current byte
reg [7:0]	dataReg;			// current byte (ready to be internally written or sent to master)
reg [7:0]	pageReg[0:63];			// Page Block (fpr "PageWrite" operations)

reg[15:0] addr, manAddr, manAddr1;
reg[5:0] manAddr3, manAddr4;
reg readA2, readA1, readA0;
reg regToBus;

reg ACKsent;

integer fMAX, tBUF, tSU_STA, tHD_STA, tSU_STO, tSU_DAT, tHD_DAT, tDH, tAA, tWR;
integer pageCounter, bitCounter, i, allMemLength, protectType, writeAllowed, tHALF_MIN, tMID_MIN;
integer seqCounter, addrBytesCounter;
integer semForever, semInternalWriting, semSoftProtect, alreadySoftProtected;
integer TestPhase,addrBytes;
integer lPage;					 
integer lSeq;	
integer slaveAddrBits, wrongHeader, wrongSlaveAddress;
integer man,man0;

time savTim;
time negSDATim;
time posSDATim;
time myTimDif1,myTimDif2;
time negSCLTim;
time posSCLTim;
time lastStopTim;

//-----------------------------------------------------------------------

tri1 SDA;
assign SDA = (busState == `OUT) ? regToBus:1'bz;

//-----------------------------------------------------------------------

`include "MemType.v"
//-----------------

//////////////////////////////////////////// TASK SECTION //////////////////////////////

task setIdleState;
	begin
	state = `idle;
	nextState = `idle;
	condition = `nothingSpecial;
	operation = `noop;
	savOperation = `noop;
	busState = `HIGH;
	bitCounter = 0;
	pageCounter = 0;
	semSoftProtect = 0;
	end
endtask

task incMemAddForRead;
	begin
	Saddr = Saddr + 1;
	#0 if(Saddr >= allMemLength)
		Saddr = 0;
	end
endtask


//////////////////////////////////////////// INITIAL SECTION //////////////////////////////

initial
	begin 

	#1  

//$monitor("====== state = %b;  bitCounter = %d",state,bitCounter,$time);

	realAddrDev = {A2,A1,A0};
	addrDev = {A2,A1,A0};
	
	`include "MemParameters.v"
	//------------------------

	tHALF_MIN = 1000000/(2*fMAX); 
	tMID_MIN = (3/8) * tHALF_MIN;
 
	//
	// memory initialization with ff
	//
	for(i=0; i<32768; i=i+1)
		mem[i] = 8'hff;

	$display;
	$display;
	#1 $display("*===============================================================");
	$display("*");
	$display("*             Chip: %s",chipName);
	$display("*");

	manAddr = allMemLength;
	manAddr1 = manAddr - 1;
	$display("* Memory size:              %d bytes (0 h - %h h)",manAddr,manAddr1);
	manAddr3 = addrBytes;
	manAddr4 = `addrBits;
	$display("* Address configuration:       %d bytes ( %d significant bits)",manAddr3, manAddr4);
	manAddr = lPage;
	$display("* Page size:                %d bytes",manAddr);

	man = fMAX;
	`ifdef standard_i2c 
		$display("* Protocol:                   standard IIC");
	`endif
	`ifdef fast_i2c
		$display("* Protocol:                   fast IIC");
	`endif
	$display("* Maximum frequency:  %d kHz",man);

	addrMax = allMemLength - 1;

	$display("*");

	case(protectType)
	`NONE:
		$display("* Write protection type:     None");

	`HALF_UPPER:
		begin
		addrBegProt = (addrMax + 1) >> 1;
		addrEndProt = addrMax;
		$display("* Write protection type:     Half upper ( %h - %h )",addrBegProt,addrEndProt);
		end

	`FULL:
		begin
		addrBegProt = 16'h0000;
		addrEndProt = addrMax;
		$display("* Write protection type:     Full ( %h - %h )",addrBegProt,addrEndProt);
		end
	endcase

	$display("*");


	case(slaveAddrBits)
		3:
			$display("*     Real device address:   %b (A2,A1,A0)",realAddrDev[2:0]);
		2:
			$display("*     Real device address:   %b (A2,A1)",realAddrDev[2:1]);
		1:
			$display("*     Real device address:   %b (A2)",realAddrDev[2]);
	endcase

	Saddr = 16'h0000;

	negSDATim = 0;
	posSDATim = 0;
	negSCLTim = 0;
	posSCLTim = 0;

	setIdleState;
	alreadySoftProtected = `N;

	busState = `HIGH;

	afterAckState = `receivingAddr;

	semInternalWriting = 0;
	lastStopTim = 0;
	ACKsent = `N;
	end


/////////////////////////////////////////// ALWAYS SECTION /////////////////////////////////

//_________________________________________________________________________________________
//
//	START detection
//

always @(negedge SDA) 
	negSDATim = $time;

always @(posedge SDA) 
	posSDATim = $time;

always @(negedge SCL) 
	negSCLTim = $time;

always @(posedge SCL) 
	posSCLTim = $time;

always @(negedge SCL &&  SDA == `B0 )
	begin
	//$display("___ Slave: SCL - (det Start) ",$time);
	myTimDif1 = negSDATim - posSCLTim;
	myTimDif2 = $time - negSDATim ;
	if( posSDATim < negSDATim && posSCLTim < negSDATim && myTimDif1 >= tSU_STA && myTimDif2 >= tHD_STA )
		begin
		if( (negSDATim - lastStopTim) < tBUF)
			$display("***** Slave (i2c memory):  START ignored (STOP - START interval < tBUF !) ");
		if(state != `internalWriting && ((negSDATim - lastStopTim) >= tBUF) )
			begin
			 $display("***** Slave (i2c memory):  START received at time ",negSDATim);
			condition = `startReceived;
			bitCounter = 0;
			nextState = `receivingAddrInfo;
			operation = `write;
			savOperation = `write;
			state = `receivingAddrInfo;
			end
		else
			if(state == `internalWriting)
				$display("***** Slave (i2c memory):  START ignored (internal writing not finished yet !) ");
		end
	else
		begin
		if( posSDATim < negSDATim && posSCLTim < negSDATim && myTimDif1 < tSU_STA )
			begin
			$display("***** Slave (i2c memory):  START ignored ");
			$display("      (tSU_STA is %d and must be >= %d) ",myTimDif1,tSU_STA,$time);
			end
		if( posSDATim < negSDATim && posSCLTim < negSDATim && myTimDif2 < tHD_STA )
			begin
			$display("***** Slave (i2c memory):  START ignored ");
			$display("      (tHD_STA is %d and must be >= %d !)",myTimDif2,tHD_STA,$time );
			end
		end
	end


//_________________________________________________________________________________________
//
//	STOP detection
//
always @(posedge SDA && state != `idle ) 
   begin
   //$display("___ Slave: SDA + (det Stop) ",$time);
   if( (posSCLTim > negSCLTim) && ($time - posSCLTim) >= tSU_STO )
	begin
	if(state != `internalWriting)
		begin
		$display("* Slave (i2c memory):  STOP received ",$time);
		condition = `stopReceived;
		lastStopTim = $time;
		case(operation)
		`noop:
			begin	
			setIdleState;
			busState = `HIGH;	
			end	
		`read:
			begin	
			setIdleState;
			busState = `HIGH;	
			end	
		`write:
			begin	
			if(writeAllowed == `Y && ACKsent == `Y) 
				begin
				$display("***** Slave (i2c memory):  Internal Writing at time ",$time);
				state = `internalWriting;
				semInternalWriting = 1;
				nextState = `internalWriting;
				busState = `HIGH;
				bitCounter = 0;
				// $display(" before WRITE: Saddr = %h, mem[Saddr] = %h, WP = %d",Saddr,mem[Saddr], WP); 
				//if(writeAllowed == `Y) 
					//begin
				for (i=0; i< pageCounter; i=i+1 )
					begin
					mem[Saddr] = pageReg[i];
					//$display(" after WRITE: Saddr = %h, mem[Saddr] = %h, WP = %d",Saddr,mem[Saddr], WP); 
					#0 Saddr[`pgExpMinusOne:0] = Saddr[`pgExpMinusOne:0] + 1;//(inc.adr.in pag)
					end	
					//end		
				condition = `nothingSpecial;
				pageCounter = 0;

				#tWR setIdleState;

				busState = `HIGH;
				bitCounter = 0;
				operation = `noop;
				savOperation = `noop;
				semInternalWriting = 0;
				$display("***** Slave (i2c memory):  end of Internal Writing at time ", $time);
				end
			else
				begin
				condition = `nothingSpecial;
				pageCounter = 0;
				setIdleState;
				busState = `HIGH;
				bitCounter = 0;
				operation = `noop;
				savOperation = `noop;
				semInternalWriting = 0;
				end
			end
		endcase
		end
	else
		lastStopTim = $time;
	end
   else
   	if( (posSCLTim > negSCLTim) && ($time - posSCLTim) < tSU_STO )
			$display("***** Slave (i2c memory): Stop Condition Setup Time < tSU_STO ", $time);

   end

//____________________________________________________________________________________________ SCL --------

always @(negedge SCL && state != `idle) 
	begin
	//$display("___ Slave: SCL - (send ACK) ",$time);
	if(state != `internalWriting)
		begin
		//
		//	ACK sending
		//
		if(state == `receivingAddrInfo && bitCounter == 9)
			begin
			busState = `HIGH;
			//$display("***** Slave (i2c memory):  After ACK at time ",$time);
			ACKsent = `Y;
			bitCounter = 0;
			if(savOperation == `read)
				afterAckState = `sendingData;
			else
				afterAckState = `receivingAddr;
			nextState = afterAckState;
			end

		if(state == `receivingAddrInfo && bitCounter == 8)
			begin
			//
			// adrInfo verification
			//
			wrongHeader = `N;
			wrongSlaveAddress = `N;
			if( tempReg[3:0] != reqHeaderFirstHalf[3:0])
				wrongHeader = `Y;

			case(slaveAddrBits)
				3:
				if( ! (readA2 == A2 && readA1 == A1 && readA0 == A0 ))
					wrongSlaveAddress = `Y;
				2:
				if( ! (readA2 == A2 && readA1 == A1) )
					wrongSlaveAddress = `Y;
				1:
				if( ! (readA2 == A2) )
					wrongSlaveAddress = `Y;
				0:
					wrongSlaveAddress = `N;
			endcase

			`ifdef cat24c164
				wrongHeader = `N;
				wrongSlaveAddress = `N;
				if( !(tempReg[3]==`B1 && tempReg[2]==A2 && tempReg[1]==~A1 && tempReg[0]==A0) )
					begin
						wrongHeader = `Y;
						wrongSlaveAddress = `Y;
					end
			`endif

			if(wrongHeader == `N &&  wrongSlaveAddress == `N )
				begin

				extraAddrBits = 0;
				nrExtraAddrBits = 0;
				if(addrBytes == 1 && `addrBits > 8)
					begin
						nrExtraAddrBits = `addrBits-8;
						case(nrExtraAddrBits)
							3:
								begin
								extraAddrBits[2]=readA2;
								extraAddrBits[1]=readA1;
								extraAddrBits[0]=readA0;
								end
							2:
								begin
								extraAddrBits[1]=readA1;
								extraAddrBits[0]=readA0;
								end
							1:
								begin
								extraAddrBits[0]=readA0;
								end
						endcase
					end

				//$display("***** Slave (i2c memory):  sending ACK at time ",$time);
				busState = `OUT;
				regToBus =`B0;	
				end
			else
				begin
				if( wrongSlaveAddress == `Y )
					$display("***** Slave (i2c memory):  unrecognized device address !");
				if( wrongHeader == `Y )
					$display("***** Slave (i2c memory):  wrong header ! (%b instead of %b)",tempReg[3:0],reqHeaderFirstHalf[3:0]);
				#tAA
				busState = `HIGH;		// no ACK !
				setIdleState;
				end
			#(tHALF_MIN) operation = savOperation;
			end

		if(state == `receivingAddr && bitCounter == 8)
			begin
			if(addrBytesCounter == 2)
				afterAckState = `receivingAddr;
			else
				afterAckState = `receivingData;

				//$display("***** Slave (i2c memory):  sending ACK at time ",$time);
			ACKsent = `Y;
			busState = `OUT;
			regToBus =`B0;
			end

		if(state == `receivingAddr && bitCounter == 9)
			begin
			busState = `HIGH;
			bitCounter = 0;
			nextState = afterAckState;
			if( WP == `B0 || ( WP == `B1 && (Saddr < addrBegProt || Saddr > addrEndProt) ) ) 
				begin
				writeAllowed = `Y;
				end
			else
				writeAllowed = `N;
			//$display("      Slave (i2c memory):  ======= received address = %h",Saddr[15:0],$time);
			end


		if(state == `receivingData && bitCounter == 8)
		begin
			if(writeAllowed == `Y || semSoftProtect == 1) 
				//
				//	Write enabled ==> ACK
				//	
				begin
				//$display("      Slave (i2c memory):  received data = %b",tempReg[7:0]);
				//$display("*********** Slave (i2c memory):  sending ACK ",$time);
				ACKsent = `Y;
				busState = `OUT;
				regToBus =`B0;
				end
			else
				if(semSoftProtect == 0)
					//
					//	Write Protection ==> no ACK
					//
					begin
					//#tAA $display("      Slave (i2c memory):  received data = %b",tempReg[7:0]);
					$display("***** Slave (i2c memory):  WRITE PROTECTION ");
					$display("***** Slave (i2c memory):  sending NO ACK at time ",$time);
					busState = `HIGH;
					end
			semSoftProtect = 0;
		end

		if(state == `receivingData && bitCounter == 9)
		begin
			if(writeAllowed == `Y) 
				begin
				bitCounter = 0; 
				#tHALF_MIN 
				busState = `HIGH;
				bitCounter = 0;
				dataReg[7:0] = tempReg[7:0];
				pageReg[pageCounter] = tempReg[7:0];
				pageCounter = pageCounter + 1;
				afterAckState = `receivingData;
				nextState = afterAckState;
				end
			else
				begin
				#tHALF_MIN nextState = afterAckState;
				#(tHALF_MIN+tDH-tAA) busState = `HIGH;	
				busState = `HIGH;
				bitCounter = 0;
				operation = `noop;	 
				nextState = afterAckState;
				end
		end
	end
end 

//_________________________________________________________________________________________ SCL -----

always @(negedge SCL && state != `idle ) 
	//
	//	State machine for SEND type Slave (i2c memory):  activity 
	//		Clock negative edge actions 
	begin
	//$display("___ Slave: SCL - (send) ",$time);
#0	if(savOperation != `write && state != `internalWriting)
		begin
		case(nextState)
		`sendingData:
			begin
			//$display("      Slave (i2c memory):  sending data ");
			case(bitCounter)
				0:
					begin
					bitCounter = bitCounter + 1;
					tempReg[7:0] = mem[Saddr];
						 //$display("============ Address: %h Byte to be read %h ",Saddr,mem[Saddr]);
					#0 
					regToBus = tempReg[7];
					if(regToBus == `B0)
						busState = `OUT;
					else
						busState = `HIGH;
					#0 tempReg = tempReg << 1;	
					end
				8:
					begin
					bitCounter = bitCounter+1;
					#tDH regToBus = tempReg[7];
					if(regToBus == `B0)
						busState = `OUT;
					else
						busState = `HIGH;
					state = `waitingAck;
					nextState = `waitingAck;
					busState = `HIGH;
					afterAckState = `sendingData;
					incMemAddForRead;
					end
				default:
					begin
					bitCounter = bitCounter + 1;
					#tDH regToBus = tempReg[7];
					if(regToBus == `B0)
						busState = `OUT;
					else
						busState = `HIGH;
					#0 tempReg = tempReg << 1;
					end
			endcase
			end

		default:
			begin
			end
		endcase
		end
	end

//_________________________________________________________________________________________ SCL +++++++
always @(posedge SCL && state != `idle ) 
	//
	//	State machine for RECEIVE type Slave (i2c memory):  activity 
	//		Clock positive edge actions 
	//
	begin
	//$display("___ Slave: SCL + (det ACK) ",$time);
	if(operation == `read && bitCounter == 9) 
	//
	//	ACK detection
	//
	#tMID_MIN if(SDA == `B0)
		begin
 		condition = `ackReceived;
		//$display("***** Slave (i2c memory):  ACK received at time ",$time);
		//	This means more data !!! (sequential operation)
		nextState = afterAckState;
		condition = `nothingSpecial;
		bitCounter = 0;
		end
	else
		begin
		$display("***** Slave (i2c memory):  No ACK received at time ",$time);
		operation = `noop;
		busState = `HIGH;
		nextState = `waitingStop;
		#tHALF_MIN bitCounter = 0;
		end
	end

//_________________________________________________________________________________________ SCL +++++++

always @(posedge SCL && state != `idle ) 
	//
	//	State machine for RECEIVE type Slave (i2c memory):  activity 
	//		Clock positive edge actions 
	//
	begin
	// ### $display("___ Slave: SCL + (rec) ",$time);
	#1
	if(state != `internalWriting)
		state = nextState;
	case(state)
		`receivingAddrInfo: 
			begin
			condition = `nothingSpecial;

			if(bitCounter < 9)
				bitCounter = bitCounter + 1;
			 //$display("* Slave (i2c memory):  receiving addrInfo %d: %v",bitCounter,SDA,$time);

			#0 case(bitCounter)
				5:	
				begin
				nextState = state;
				#tMID_MIN readA2 = SDA;
				end		
				6:	
				begin
				nextState = state;
				#tMID_MIN readA1 = SDA;					
				end				
				7:
				begin
				nextState = state;
				#tMID_MIN readA0 = SDA;
				end	
				8:
				begin
				#tMID_MIN if(SDA == `B1) 
					begin
					#tMID_MIN 
					savOperation = `read; // Read com. from master
					operationType = `immedAddrRead;
					afterAckState = `sendingData;
					nextState = state;
					//nextState = `sendingData;
					end
				else 
					begin
					savOperation = `write;
					operation = `write;// Write com. from master 
						// (or dummy write for selective read)
					nextState = `receivingAddrInfo;
					afterAckState = `receivingAddr;
					end
				end
				9:
				begin	
				addrBytesCounter = addrBytes;
				end
				1:
				begin
				ACKsent = `N;
				tempReg = 8'b00000000;
				#delta tempReg[0] = SDA;
				nextState = `receivingAddrInfo;
				end
				// cases 2,3,4
				default:
				begin
				nextState = `receivingAddrInfo;
				tempReg = tempReg << 1;
				#tMID_MIN tempReg[0] = SDA;
		 		end
			endcase
			end
				
		`receivingAddr:
			begin
			//$display("      Slave (i2c memory):  receiving address");
			bitCounter = bitCounter + 1;
			#0 case(bitCounter) 
				9:
					begin
					// nextState = `receivingAddr;
					#tMID_MIN 
					case(addrBytesCounter)
					2:
						begin
						nextState = `receivingAddr;
						addrBytesCounter = addrBytesCounter-1;
						Saddr[15:8] = tempReg[7:0];
						//$display("            SLAVE:  rec. addr[15:8]:             %h",tempReg[7:0]);
						end
					1:
						begin
						Saddr[7:0] = tempReg[7:0];
						
						if(addrBytes == 1)
							begin
							Saddr[15:11] = 0;
							Saddr[10:8] = extraAddrBits[2:0];
							end

						if(operation == `read) 
							begin
							nextState = `sendingData;
							#(tHALF_MIN) bitCounter = 0;
							end
						else
							begin
							nextState = `receivingData;
							end
						end
					endcase
					end
				1:
					begin
					ACKsent = `N;
					nextState = `receivingAddr;
					tempReg = 8'b00000000;
					#tMID_MIN tempReg[0] = SDA;
					end
				// cases 2,3,4,5,6,7
				default:
					begin
					nextState = `receivingAddr;
					tempReg = tempReg << 1;
					#tMID_MIN tempReg[0] = SDA;
		 			end
				endcase

			end

		`receivingData :
			begin
			//$display("      Slave (i2c memory):  receiving first data");
			// ### bitCounter = bitCounter + 1;

			#1 bitCounter = bitCounter + 1;
			case(bitCounter)

			9: 
				begin
				end
			1:
				begin
				tempReg = 8'b00000000;
				if(condition != `startReceived)
					#tMID_MIN tempReg[0] = SDA;
				else
					begin
					operation = `read;
					operationType = `selectiveRead;
					nextState = `receivingAddrInfo;
					end	
				end
			// cases 2,3,4,5,6,7,8
			default:
				begin
				ACKsent = `N;
				tempReg = tempReg << 1;
				#tMID_MIN tempReg[0] = SDA;
				end
			endcase
			end

		`waitingAck:
			begin
			nextState = afterAckState;
			#(tHALF_MIN+tMID_MIN) state = `sendingData;
			end

		`waitingStop:
			begin
			$display("      Slave (i2c memory):  waiting STOP");
			nextState = state;
			end

		`internalWriting:
			begin
				#(5*tHALF_MIN) nextState = `receivingAddrInfo;
			end
	endcase 
	end  
endmodule
